"""
Extension for the `FeedCog`.
Loading this file via `commands.Bot.load_extension` will add `FeedCog` to the bot.
"""

import logging
from typing import Tuple

import validators
from feedparser import FeedParserDict, parse
from discord.ext import commands
from discord import Interaction, Embed, Colour, TextChannel, Permissions
from discord.app_commands import Choice, Group, autocomplete, choices, rename
from sqlalchemy import insert, select, and_, delete
from sqlalchemy.exc import NoResultFound

from utils import get_rss_data, followup, audit, followup_error  # pylint: disable=E0401
from feed import get_source, get_unparsed_feed, Source  # pylint: disable=E0401
from db import (  # pylint: disable=E0401
    DatabaseManager,
    SentArticleModel,
    RssSourceModel,
    FeedChannelModel,
    AuditModel
)

log = logging.getLogger(__name__)

rss_list_sort_choices = [
    Choice(name="Nickname", value=0),
    Choice(name="Date Added", value=1)
]
channels_list_sort_choices=[
    Choice(name="Feed Nickname", value=0),
    Choice(name="Channel ID", value=1),
    Choice(name="Date Added", value=2)
]

# TODO SECURITY: a potential attack is that the user submits an rss feed then changes the
# target resource. Run a period task to check this.
async def validate_rss_source(nickname: str, url: str) -> Tuple[str | None, FeedParserDict | None]:
    """Validate a provided RSS source.

    Parameters
    ----------
    nickname : str
        Nickname of the source. Must not contain URL.
    url : str
        URL of the source. Must be URL with valid status code and be an RSS feed.

    Returns
    -------
    str or None
        String invalid message if invalid, NoneType if valid.
    FeedParserDict or None
        The feed parsed from the given URL or None if invalid.
    """

    # Ensure the URL is valid
    if not validators.url(url):
        return f"The URL you have entered is malformed or invalid:\n`{url=}`", None

    # Check the nickname is not a URL
    if validators.url(nickname):
        return "It looks like the nickname you have entered is a URL.\n" \
                f"For security reasons, this is not allowed.\n`{nickname=}`", None


    feed_data, status_code = await get_rss_data(url)

    # Check the URL status code is valid
    if status_code != 200:
        return f"The URL provided returned an invalid status code:\n{url=}, {status_code=}", None

    # Check the contents is actually an RSS feed.
    feed = parse(feed_data)
    if not feed.version:
        return f"The provided URL '{url}' does not seem to be a valid RSS feed.", None

    return None, feed

async def set_all_articles_as_sent(inter, channel: TextChannel, feed_id: int, rss_url: str):
    unparsed_feed = await get_unparsed_feed(rss_url)
    source = Source.from_parsed(parse(unparsed_feed))
    articles = source.get_latest_articles()

    async with DatabaseManager() as database:
        query = insert(SentArticleModel).values([
            {
                "discord_server_id": inter.guild_id,
                "discord_channel_id": channel.id,
                "discord_message_id": -1,
                "article_url": article.url,
                "feed_channel_id": feed_id 
            }
            for article in articles
        ])
        await database.session.execute(query)


class FeedCog(commands.Cog):
    """
    Command cog.
    """

    def __init__(self, bot):
        super().__init__()
        self.bot = bot

    @commands.Cog.listener()
    async def on_ready(self):
        """Instructions to call when the cog is ready."""

        log.info("%s cog is ready", self.__class__.__name__)

    async def source_autocomplete(self, inter: Interaction, nickname: str):
        """Provides RSS source autocomplete functionality for commands.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        nickname : str
            _description_

        Returns
        -------
        list of app_commands.Choice
            _description_
        """

        async with DatabaseManager() as database:
            whereclause = and_(
                RssSourceModel.discord_server_id == inter.guild_id,
                RssSourceModel.nick.ilike(f"%{nickname}%")
            )
            query = select(RssSourceModel).where(whereclause).order_by(RssSourceModel.nick)
            result = await database.session.execute(query)
            sources = [
                Choice(name=rss.nick, value=rss.rss_url)
                for rss in result.scalars().all()
            ]

        return sources

    # All RSS commands belong to this group.
    feed_group = Group(
        name="feed",
        description="Commands for rss sources.",
        guild_only=True,  # We store guild IDs in the database, so guild only = True
        default_permissions=Permissions.elevated()
    )

    @feed_group.command(name="add")
    async def add_rss_source(self, inter: Interaction, nickname: str, url: str):
        """Add a new RSS source.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        nickname : str
            A name used to identify the RSS source.
        url : str
            The RSS feed URL.
        """

        await inter.response.defer()

        illegal_message, feed = await validate_rss_source(nickname, url)
        if illegal_message:
            await followup(inter, illegal_message, suppress_embeds=True)
            return

        log.debug("RSS feed added")

        async with DatabaseManager() as database:
            query = insert(RssSourceModel).values(
                discord_server_id = inter.guild_id,
                rss_url = url,
                nick=nickname
            )
            await database.session.execute(query)

            await audit(self,
                f"Added RSS source ({nickname=}, {url=})",
                inter.user.id, database=database
            )

        embed = Embed(title="RSS Feed Added", colour=Colour.dark_green())
        embed.add_field(name="Nickname", value=nickname)
        embed.add_field(name="URL", value=url)
        embed.set_thumbnail(url=feed.get("feed", {}).get("image", {}).get("href"))

        await followup(inter, embed=embed)

    @feed_group.command(name="remove")
    @rename(url="option")
    @autocomplete(url=source_autocomplete)
    async def remove_rss_source(self, inter: Interaction, url: str):
        """Delete an existing RSS source.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        url : str
            The RSS source to be removed. Autocomplete or enter the URL.
        """

        await inter.response.defer()

        log.debug("Attempting to remove RSS source (url=%s)", url)

        async with DatabaseManager() as database:
            whereclause = and_(
                RssSourceModel.discord_server_id == inter.guild_id,
                RssSourceModel.rss_url == url
            )

            # We will select the item first, so we can reference it's nickname later.
            select_query = select(RssSourceModel).filter(whereclause)
            select_result = await database.session.execute(select_query)

            try:
                rss_source = select_result.scalars().one()
            except NoResultFound:
                await followup_error(inter,
                    title="Error Deleting Feed",
                    message=f"I couldn't find anything for `{url}`"
                )
                return

            nickname = rss_source.nick

            delete_query = delete(RssSourceModel).filter(whereclause)
            delete_result = await database.session.execute(delete_query)

            await audit(self,
                f"Deleted RSS source ({nickname=}, {url=})",
                inter.user.id, database=database
            )

        source = get_source(url)  # TODO: replace with async function

        embed = Embed(title="RSS Feed Deleted", colour=Colour.dark_red())
        embed.add_field(name="Nickname", value=nickname)
        embed.add_field(name="URL", value=url)
        embed.set_thumbnail(url=source.icon_url)

        await followup(inter, embed=embed)

    @feed_group.command(name="list")
    @choices(sort=rss_list_sort_choices)
    async def list_rss_sources(
        self, inter: Interaction, sort: Choice[int]=None, sort_reverse: bool=False
    ):
        """Provides a with a list of RSS sources available for the current server.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        """

        await inter.response.defer()

        # Default to the first choice if not specified.
        if isinstance(sort, Choice):
            description = "Sort by "
            description += "Nickname " if sort.value == 0 else "Date Added "
            description += '\U000025BC' if sort_reverse else '\U000025B2'
        else:
            sort = rss_list_sort_choices[0]
            description = ""

        match sort.value, sort_reverse:
            case 0, False:
                order_by = RssSourceModel.nick.asc()
            case 0, True:
                order_by = RssSourceModel.nick.desc()
            case 1, False:
                order_by = RssSourceModel.created.desc()
            case 1, True:
                order_by = RssSourceModel.created.asc()
            case _, _:
                raise ValueError(f"Unknown sort: {sort}")

        async with DatabaseManager() as database:
            whereclause = and_(RssSourceModel.discord_server_id == inter.guild_id)
            query = select(RssSourceModel).where(whereclause).order_by(order_by)
            result = await database.session.execute(query)

            rss_sources = result.scalars().all()
            rowcount = len(rss_sources)

            if not rss_sources:
                await followup_error(inter,
                    title="No Feeds Found",
                    message="I couldn't find any Feeds for this server."
                )
                return

            output = "\n".join([
                f"{i}. **[{rss.nick}]({rss.rss_url})** "
                for i, rss in enumerate(rss_sources)
            ])

        embed = Embed(
            title="Saved RSS Feeds",
            description=f"{description}\n\n{output}",
            colour=Colour.blue()
        )
        embed.set_footer(text=f"Showing {rowcount} results")

        await followup(inter, embed=embed)

    # @feed_group.command(name="fetch")
    # @rename(max_="max")
    # @autocomplete(rss=source_autocomplete)
    async def fetch_rss(self, inter: Interaction, rss: str, max_: int=1):
        """Fetch an item from the specified RSS feed.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        rss : str
            The RSS feed to fetch from.
        max_ : int, optional
            Maximum number of items to fetch, by default 1, limits at 5.
        """

        await inter.response.defer()

        if max_ > 5:
            followup(inter, "It looks like you have requested too many articles.\nThe limit is 5")
            return

        invalid_message, feed = await validate_rss_source("", rss)
        if invalid_message:
            await followup(inter, invalid_message)
            return

        source = Source.from_parsed(feed)
        articles = source.get_latest_articles(max_)

        if not articles:
            await followup(inter, "Sorry, I couldn't find any articles from this feed.")
            return

        embeds = [await article.to_embed() for article in articles]

        async with DatabaseManager() as database:
            query = insert(SentArticleModel).values([
                {
                    "discord_server_id": inter.guild_id,
                    "discord_channel_id": inter.channel_id,
                    "discord_message_id": inter.id,
                    "article_url": article.url,
                }
                for article in articles
            ])
            await database.session.execute(query)
            await audit(self,
                f"User is requesting {max_} articles from {source.name}",
                inter.user.id, database=database
            )

        await followup(inter, embeds=embeds)

    # Help     ---- ---- ----

    @feed_group.command(name="help")
    async def get_help(self, inter: Interaction):
        """Get help on how to use my commands.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        """

        await inter.response.defer()

        description = (
            "`/feed add <nickname> <url>` \n\n"
            "Save a new RSS feed to the bot. This can be referred to later, when assigning "
            "channels to receive content from these RSS feeds."

            "\n\n\n`/feed remove <option>` \n\n"
            "Remove a previously saved RSS feed. Select the nickname from the shown options "
            "if any. You can re-add the RSS feed to the bot using the `/feeds add` command."

            "\n\n\n`/feed list <sort> <sort_reverse>` \n\n"
            "List all saved RSS feeds numerically. Use the `<sort>` option to order "
            "the results by either nickname or date & time added. Use the `<sort_reverse>` "
            "option to order by ascending or descending in conjunction with the `<sort>` option."

            "\n\n\n`/feed assign <rss> <textchannel>` \n\n"
            "Assign a channel to an RSS feed. Previously saved RSS feeds will be selectable "
            "under the `<rss>` option. The channel will be assumed as the current channel, "
            "unless specified otherwise using the `<textchannel` option."

            "\n\n\n`/feed unassign <option>` \n\n"
            "Unassigned channel from an RSS feed. Previously assigned channels will be shown "
            "as an `<option>`, select one to remove it."

            "\n\n\n`/feed channels` \n\n"
            "List all channels assigned to an RSS feed numerically."
        )

        embed = Embed(
            title="Help",
            description=description,
            colour=Colour.blue(),
        )

        await followup(inter, embed=embed)

    # Channels ---- ---- ----

    async def autocomplete_rss_sources(self, inter: Interaction, nickname: str):
        """_summary_

        Parameters
        ----------
        inter : Interaction
            _description_
        nickname : str
            _description_

        Returns
        -------
        _type_
            _description_
        """

        async with DatabaseManager() as database:
            whereclause = and_(
                RssSourceModel.discord_server_id == inter.guild_id,
                RssSourceModel.nick.ilike(f"%{nickname}%")
            )
            query = select(RssSourceModel).where(whereclause)
            result = await database.session.execute(query)
            sources = [
                Choice(name=rss.nick, value=rss.id)
                for rss in result.scalars().all()
            ]

        log.debug("Autocomplete rss_sources returned %s results", len(sources))

        return sources

    async def autocomplete_existing_feeds(self, inter: Interaction, current: str):
        """Returns a list of existing RSS + Channel feeds.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        current : str
            The current text entered for the autocomplete.
        """

        async with DatabaseManager() as database:
            whereclause = and_(
                FeedChannelModel.discord_server_id == inter.guild_id,
                FeedChannelModel.search_name.ilike(f"%{current}%"),
                RssSourceModel.id == FeedChannelModel.rss_source_id
            )
            query = (
                select(FeedChannelModel, RssSourceModel)
                .where(whereclause)
                .join(RssSourceModel)
                .order_by(FeedChannelModel.discord_channel_id)
            )
            result = await database.session.execute(query)
            feeds = []
            for feed in result.scalars().all():
                channel = inter.guild.get_channel(feed.discord_channel_id)
                feeds.append(Choice(
                    name=f"# {channel.name}  |  {feed.rss_source.nick}",
                    value=feed.id
                ))

        log.debug("Autocomplete existing_feeds returned %s results", len(feeds))

        return feeds

    # # All RSS commands belong to this group.
    # channel_group = Group(
    #     name="channels",
    #     description="Commands for channel assignment.",
    #     guild_only=True  # These commands belong to channels of
    # )

    @feed_group.command(name="assign")
    @rename(rss="feed")
    @autocomplete(rss=autocomplete_rss_sources)
    async def include_feed(
        self, inter: Interaction, rss: int, channel: TextChannel = None, prevent_spam: bool = True
    ):
        """Include a feed within the specified channel.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        rss : int
            The RSS feed to include.
        channel : TextChannel
            The channel to include the feed in.
        """

        await inter.response.defer()

        channel = channel or inter.channel

        async with DatabaseManager() as database:
            select_query = select(RssSourceModel).where(and_(
                RssSourceModel.id == rss,
                RssSourceModel.discord_server_id == inter.guild_id
            ))

            select_result = await database.session.execute(select_query)
            rss_source = select_result.scalars().one()
            nick, rss_url = rss_source.nick,  rss_source.rss_url

            insert_query = insert(FeedChannelModel).values(
                discord_server_id = inter.guild_id,
                discord_channel_id = channel.id,
                rss_source_id=rss,
                search_name=f"{nick} #{channel.name}"
            )

            insert_result = await database.session.execute(insert_query)
            feed_id = insert_result.inserted_primary_key.id

        if prevent_spam:
            await set_all_articles_as_sent(inter, channel, feed_id, rss_url)

        await followup(inter, f"I've included [{nick}]({rss_url}) to {channel.mention}")

    @feed_group.command(name="unassign")
    @autocomplete(option=autocomplete_existing_feeds)
    async def exclude_feed(self, inter: Interaction, option: int):
        """Undo command for the `/channel include-feed` command.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        option : str
            The RSS feed and channel to exclude.
        """

        await inter.response.defer()

        async with DatabaseManager() as database:
            query = delete(FeedChannelModel).where(and_(
                FeedChannelModel.id == option,
                FeedChannelModel.discord_server_id == inter.guild_id
            ))

            result = await database.session.execute(query)

        if not result.rowcount:
            await followup_error(inter,
                title="Assigned Feed Not Found",
                message=f"I couldn't find any assigned feeds for the option: {option}"
            )
            return

        await followup(inter, "I've removed this item (placeholder response)")

    @feed_group.command(name="channels")
    @choices(sort=channels_list_sort_choices)
    async def list_feeds(self, inter: Interaction, sort: Choice[int] = 0, sort_reverse: bool = False):
        """List all of the channels and their respective included feeds.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        """

        await inter.response.defer()

        description = "Sort By "

        if isinstance(sort, Choice):
            match sort.value, sort_reverse:
                case 0, False:
                    order_by = RssSourceModel.nick.asc()
                    description += "Nickname "
                case 0, True:
                    order_by = RssSourceModel.nick.desc()
                    description += "Nickname "
                case 1, False:
                    order_by = FeedChannelModel.discord_channel_id.asc()
                    description += "Channel ID "
                case 1, True:
                    order_by = FeedChannelModel.discord_channel_id.desc()
                    description += "Channel ID "
                case 2, False:
                    order_by = RssSourceModel.created.desc()
                    description += "Date Added "
                case 2, True:
                    order_by = RssSourceModel.created.asc()
                    description += "Date Added "
                case _, _:
                    raise ValueError(f"Unknown sort: {sort}")
        else:
            order_by = FeedChannelModel.discord_channel_id.asc()
            description = ""

        async with DatabaseManager() as database:
            whereclause = and_(
                FeedChannelModel.discord_server_id == inter.guild_id,
                RssSourceModel.id == FeedChannelModel.rss_source_id
            )
            query = (
                select(FeedChannelModel, RssSourceModel)
                .where(whereclause)
                .join(RssSourceModel)
                .order_by(order_by)
            )
            result = await database.session.execute(query)

            feed_channels = result.scalars().all()
            rowcount = len(feed_channels)

            if not feed_channels:
                await followup_error(inter,
                    title="No Assigned Feeds Found",
                    message="Assign a channel to receive feed content with `/feed assign`."
                )
                return

            output = "\n".join([
                f"{i}. <#{feed.discord_channel_id}> · "
                f"[{feed.rss_source.nick}]({feed.rss_source.rss_url})"
                for i, feed in enumerate(feed_channels)
            ])

        embed = Embed(
            title="Saved Feed Channels",
            description=f"{description}\n{output}",
            colour=Colour.blue()
        )
        embed.set_footer(text=f"Showing {rowcount} results")

        await followup(inter, embed=embed)

    admin_group = Group(
        name="admin",
        description="Administration tasks",
        guild_only=True,
        default_permissions=Permissions.elevated()
    )

    @admin_group.command(name="clear-sent-articles")
    async def clear_sent_articles(self, inter: Interaction):
        """Clear the database of all sent articles.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        """

        await inter.response.defer()

        async with DatabaseManager() as database:
            query = delete(SentArticleModel).where(and_(
                SentArticleModel.discord_server_id == inter.guild_id
            ))
            result = await database.session.execute(query)

        await followup(inter,
            f"{result.rowcount} sent articles have been cleared from the database. "
            "I will no longer recognise these articles as sent, and will send them "
            "again if they appear during the next RSS feed scan."
        )

    audit_group = Group(
        name="audit",
        description="Check audited actions.",
        guild_only=True,
        default_permissions=Permissions.elevated()
    )

    @audit_group.command(name="check")
    async def check_audit_log(self, inter: Interaction):
        """Check the audit log.

        Parameters
        ----------
        inter : Interaction
            Represents an app command interaction.
        """

        await inter.response.defer()

        async with DatabaseManager() as database:
            query = select(AuditModel).where(and_(
            ))
            result = await database.session.execute(query)


async def setup(bot):
    """
    Setup function for this extension.
    Adds `FeedCog` to the bot.
    """

    cog = FeedCog(bot)
    await bot.add_cog(cog)
    log.info("Added %s cog", cog.__class__.__name__)
