
import json
import logging
import async_timeout
from dataclasses import dataclass
from datetime import datetime

import aiohttp
            
import validators
from textwrap import shorten
from markdownify import markdownify
from discord import Embed, Colour
from bs4 import BeautifulSoup as bs4
from feedparser import FeedParserDict, parse

log = logging.getLogger(__name__)
dumps = lambda _dict: json.dumps(_dict, indent=8)


@dataclass
class Article:
    """Represents a news article, or entry from an RSS feed."""

    title: str | None
    description: str | None
    url: str | None
    published: datetime | None
    author: str | None
    source: object

    @classmethod
    def from_entry(cls, source, entry:FeedParserDict):
        """Create an Article from an RSS feed entry.

        Parameters
        ----------
        entry : FeedParserDict
            An entry pulled from a complete FeedParserDict object.

        Returns
        -------
        Article
            The Article created from the feed entry.
        """

        log.debug("Creating Article from entry: %s", dumps(entry))

        published_parsed = entry.get("published_parsed")
        published = datetime(*entry.published_parsed[0:-2]) if published_parsed else None

        return cls(
            title=entry.get("title"),
            description=entry.get("description"),
            url=entry.get("link"),
            published=published,
            author = entry.get("author"),
            source=source
        )

    async def get_thumbnail_url(self) -> str | None:
        """Returns the thumbnail URL for an article.

        Returns
        -------
        str or None
            The thumbnail URL, or None if not found.
        """

        log.debug("Fetching thumbnail for article: %s", self)

        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(self.url) as response:
                    html = await response.text()
        except aiohttp.InvalidURL as error:
            log.error(error)
            return None

        soup = bs4(html, "html.parser")
        image_element = soup.select_one("meta[property='og:image']")
        if not image_element:
            return None

        image_content = image_element.get("content")
        return image_content if validators.url(image_content) else None

    async def to_embed(self) -> Embed:
        """Creates and returns a Discord Embed object from the article.

        Returns
        -------
        Embed
            A Discord Embed object representing the article.
        """

        log.debug(f"Creating embed from article: {self}")

        # Replace HTML with Markdown, and shorten text.
        title = shorten(markdownify(self.title, strip=["img", "a"]), 256)
        desc = shorten(markdownify(self.description, strip=["img"]), 4096)
        author = shorten(self.source.name, 256)

        # validate urls
        embed_url = self.url if validators.url(self.url) else None
        author_url = self.source.url if validators.url(self.source.url) else None
        icon_url = self.source.icon_url if validators.url(self.source.icon_url) else None
        thumb_url = await self.get_thumbnail_url()  # validation done inside func

        embed = Embed(
            title=title,
            description=desc,
            timestamp=self.published,
            url=embed_url,
            colour=Colour.blue()
        )

        embed.set_thumbnail(url=icon_url)
        embed.set_image(url=thumb_url)
        embed.set_author(url=author_url, name=author)
        embed.set_footer(text=self.author)

        return embed


@dataclass
class Source:
    """Represents an RSS source."""
    
    name: str | None
    url: str | None
    icon_url: str | None
    feed: FeedParserDict

    @classmethod
    def from_parsed(cls, feed:FeedParserDict):
        """Returns a Source object from a parsed feed.

        Parameters
        ----------
        feed : FeedParserDict
            The feed used to create the Source.

        Returns
        -------
        Source
            The Source object
        """

        log.debug("Creating Source from feed: %s", dumps(feed))

        return cls(
            name=feed.get("channel", {}).get("title"),
            url=feed.get("channel", {}).get("link"),
            icon_url=feed.get("feed", {}).get("image", {}).get("href"),
            feed=feed
        )

    def get_latest_articles(self, max: int = 999) -> list[Article]:
        """Returns a list of Article objects.

        Parameters
        ----------
        max : int
            The maximum number of articles to return.

        Returns
        -------
        list of Article
            A list of Article objects.
        """

        log.debug("Fetching latest articles from %s, max=%s", self, max)

        return [
            Article.from_entry(self, entry)
            for i, entry in enumerate(self.feed.entries)
            if i < max
        ]
    
async def fetch(session, url: str) -> str:
    async with async_timeout.timeout(20):
        async with session.get(url) as response:
            return await response.text()

async def get_unparsed_feed(url: str):
    async with aiohttp.ClientSession() as session:
        return await fetch(session, url)


def get_source(rss_url: str) -> Source:
    """_summary_

    Parameters
    ----------
    rss_url : str
        _description_

    Returns
    -------
    Source
        _description_
    """

    parsed_feed = parse(rss_url)  # TODO: make asyncronous
    return Source.from_parsed(parsed_feed)
